from WeeLab.devices.pilas import NktPilasLaser


def test_pilas_command(mock_connection):
    pilas = NktPilasLaser(mock_connection)
    # caen.set_voltage(80)
    pilas.disable()
    pilas.set_frequency(1000)
    pilas.set_tune(500)
    pilas.set_trigger_mode("internal")
    pilas.enable()

    assert mock_connection.last_command == "ld=1"



from WeeLab.devices.stage import Stage


def test_stage_command(mock_connection):
    stage = Stage(mock_connection)
    stage.move(10, 20)
    stage.ask_position()
    stage.move(1, 2)
    # caen.set_voltage_modulating_ramp(23, 20)
    assert mock_connection.last_command == "move 320 640 "



from WeeLab.connections.base import InstrumentConnection


class VisaConnection(InstrumentConnection):
    def __init__(self, resource_name: str):
        import pyvisa

        self.rm = pyvisa.ResourceManager()
        self.inst = self.rm.open_resource(resource_name)

    def write(self, command: str):
        self.inst.write(command)

    def read(self) -> str:
        return self.inst.read()



# Connections
from .connections.vxi11conn import Vxi11Connection
from .connections.visa import VisaConnection
from .connections.serial import SerialConnection

# Devices
from .devices.tek_afg import Tektronix_AFG3252
from .devices.lecroy9104 import LeCroyWaveRunner
from .devices.caen_a7585 import CAEN_A7585
from .devices.pilas import NktPilasLaser
from .devices.stage import Stage
from .manager import DeviceManager

__all__ = [
    "Vxi11Connection",
    "VisaConnection",
    "SerialConnection",
    "DeviceManager",
    "Tektronix_AFG3252",
    "LeCroyWaveRunner",
    "CAEN_A7585",
    "NktPilasLaser",
    "Stage",
]



import pytest
from WeeLab.connections.visa import InstrumentConnection


class MockConnection(InstrumentConnection):
    def __init__(self):
        self.last_command = ""

    def write(self, command: str):
        self.last_command = command

    def read(self) -> str:
        if self.last_command == "C1:WF? DAT1":
            return "MOCK_WAVEFORM_DATA"
        if self.last_command == "currpos ":
            return "currpos 1 1 "
        # return "UNKNOWN_COMMAND"
        return "OK"

    def read_all(self) -> str:
        return "MOCK_READ_ALL_DATA"

    def get_serial_number(self) -> str:
        return "MOCK_SERIAL_NUMBER"

    def close(self):
        pass

    def query(self, command: str) -> str:
        self.write(command)
        return self.read()


@pytest.fixture
def mock_connection():
    return MockConnection()



import threading
import time
import serial.tools
from WeeLab.interfaces.power_supply import PowerSupply
from WeeLab.connections.base import InstrumentConnection
from WeeLab.connections.serial import SerialConnection

from colorama import Fore, Style

"""CAEN A7585 Power Supply Control Module for SiPM Biasing.
This module provides an interface to control the CAEN A7585 power supply,
allowing for setting voltage, current, and ramp speed, as well as reading
the current and voltage values.

Example usage:
```python
from LabPack.devices.caen_a7585 import CAEN_A7585
from LabPack.connections.serial import SerialConnection

caen = CAEN_A7585(
    conn=SerialConnection(port="COM6", baudrate=115200))
caen.set_voltage(30)  # Set voltage to 30V
caen.wait_setpoint_reached()  # Wait until the setpoint is reached
caen.get_IV()  # Get the current and voltage values
"""


class CAEN_A7585(PowerSupply):
    def __init__(self, conn: InstrumentConnection, serial_number: str = "0"):
        """
        Initialize the CAEN A7585 power supply.
        :param conn: An instance of InstrumentConnection (e.g., SerialConnection).
        :param serial_number: Serial number of the CAEN A7585 device.
                              If not using InstrumentConnection, this must be provided.

        Example:
        ``` python
        from LabPack.devices.caen_a7585 import CAEN_A7585
        from LabPack.connections.serial import SerialConnection
        caen = CAEN_A7585(
            conn=SerialConnection(port="COM6", baudrate=115200))
        ```
        """

        # self.conn = conn
        self.mutex = threading.Lock()

        self.connected = False
        if isinstance(conn, InstrumentConnection):
            self.conn = conn  # Connection is provided
            self.initialize()
        else:
            if serial_number == "0":  # Neither connection nor serial number provided
                raise ValueError(
                    "Serial number must be provided if not using InstrumentConnection."
                )
            else:  # Connection is not provided, use serial number to connect
                self.connect_no_port(serial_number)

        self.serial_number = self.conn.get_serial_number()

    def initialize(self):
        self.set_machine_mode()
        self.set(0, 0)
        r = self.read_line_with_timeout(1000)
        if "OK" in r:
            self.set(0, 0)  # Turn OFF initially
            self.set(1, 0)  # digital mode
            self.set(3, 5)  # ramp [V/s]
            self.set(2, 20)  # voltage [V]
            self.set(5, 5)  # max current [mA]
            self.set(4, 80)  # max voltage [V]
            time.sleep(0.2)

            self.connected = True
            self.read_all()
            print(
                Fore.GREEN
                + "CAEN HV module A7585 biasing SiPM initialized."
                + Style.RESET_ALL
            )
        else:
            self.conn.close()
            raise Exception(
                "Error: CAEN HV module A7585 biasing SiPM could not be initialized."
            )

    def connect_no_port(self, serial_number: str):
        if not self.connected:
            ports = serial.tools.list_ports.comports()
            exception_msg_list = []
            exception_port_list = []

            available_serial_number = [
                port.serial_number for port in ports if port.serial_number is not None
            ]
            if serial_number not in available_serial_number:
                raise Exception(
                    f"Error: CAEN HV module A7585 biasing SiPM named {serial_number} seems not connected to the computer."
                    f"\nAvailable ones are : {available_serial_number}"
                )

            for port in ports:
                if serial_number:
                    if port.serial_number != serial_number:
                        continue
                try:
                    # print(port.device)
                    # self.serialPort = serial.Serial(
                    #     port=port.device,
                    #     baudrate=115200,
                    #     bytesize=8,
                    #     parity="N",
                    #     stopbits=1,
                    #     timeout=1,
                    # )
                    self.conn = SerialConnection(
                        port=port.device,
                        baudrate=115200,
                    )
                    self.serial_number = port.serial_number
                    self.initialize()
                    break
                except Exception as ex:
                    # print(f"Error connecting to port {port}: {ex}")
                    exception_msg_list.append(ex)
                    exception_port_list.append(port)
            if not self.connected:
                for port, ex in zip(exception_msg_list, exception_port_list):
                    print(f"Error connecting to port {port}: {ex}")
                raise Exception(
                    "Error: CAEN HV module A7585 biasing SiPM could not be accessed"
                )
        self.target_hv = 0

        print(f"CAEN HV {self.serial_number} connected successfully.")

    def set_machine_mode(self):
        self.conn.write("AT+MACHINE")

    def set(self, add, val):
        self.conn.write(("AT+SET," + str(add) + "," + str(val)))

    def get(self, add):
        self.conn.write(("AT+GET," + str(add)))

    def read_all(self):
        return self.conn.read_all()

    def read(self):
        return self.conn.read()

    def read_line_with_timeout(self, timeout_milliseconds):
        line_read = False
        received_line = ""

        def read_from_serial():
            nonlocal line_read, received_line
            with self.mutex:
                received_line = self.read()
            line_read = True

        serial_thread = threading.Thread(target=read_from_serial)
        serial_thread.start()
        serial_thread.join(timeout_milliseconds)
        if serial_thread.is_alive():
            line_read = False
        if line_read:
            return received_line
        else:
            return ""

    def set_enHV(self, bit):
        self.set(0, bit)
        r = self.read_line_with_timeout(1000)
        if r == "":
            print("Timeout error")

    def set_voltage(self, value_V):
        self.target_hv = value_V
        self.set(2, value_V)  # Volts
        r = self.read_line_with_timeout(1000)
        if r == "":
            print("Timeout error")
        self.set_enHV(1)
        self.set_enHV(1)

    def set_voltage_modulating_ramp(self, value_V, SiPM_breakdown):
        self.target_hv = value_V
        self.set(3, 10)  # ramp [V/s]
        if value_V < SiPM_breakdown:
            self.set_voltage(value_V)
        else:
            self.set_voltage(SiPM_breakdown)
            time.sleep(SiPM_breakdown / 10)
            self.set(3, 5)  # ramp [V/s]
            self.set_voltage((SiPM_breakdown + value_V) / 2)
            time.sleep((SiPM_breakdown + value_V) / 2 / 5)
            self.set(3, 2)  # ramp [V/s]
            self.set_voltage(value_V)

    def set_max_current(self):
        self.set(5, 1)  # mA
        r = self.read_line_with_timeout(1000)
        if r == "":
            print("Timeout error")

    def set_ramp_speed(self):
        self.set(3, 5)  # V/s
        r = self.read_line_with_timeout(1000)
        if r == "":
            print("Timeout error")

    def get_IV(self):
        hv_eff_current, hv_eff_voltage = 0, 0
        if self.connected:
            self.read_all()
            self.get(231)
            r = self.read_line_with_timeout(1000)
            if len(r) > 4:
                hv_eff_voltage = float(r[3:-3])
            self.get(232)
            r = self.read_line_with_timeout(1000)
            if len(r) > 4:
                hv_eff_current = float(r[3:-3])
        return hv_eff_current, hv_eff_voltage

    def wait_setpoint_reached(self, target_hv=None):
        iter = 0
        if target_hv is None:
            target_hv = self.target_hv
        while iter < 40:  # 20 seconds
            hv_eff_current, hv_eff_voltage = self.get_IV()
            # self.save_IV(hv_eff_current, hv_eff_voltage)

            if (hv_eff_voltage > (target_hv - 0.2)) & (
                hv_eff_voltage < (target_hv + 0.2)
            ):
                # self.timer_setpoint_reached.cancel()
                print(
                    Fore.GREEN
                    + "Setpoint reached: %.2f µA  %.1f V"
                    % (hv_eff_current * 1e3, hv_eff_voltage)
                    + Style.RESET_ALL
                )
                break
            elif iter == 19:
                print(
                    Fore.RED
                    + "Setpoint NOT reached: %.2f µA  %.1f V"
                    % (hv_eff_current * 1e3, hv_eff_voltage)
                    + Style.RESET_ALL
                )
            iter += 1
            time.sleep(0.5)



from abc import ABC, abstractmethod


class InstrumentConnection(ABC):
    @abstractmethod
    def write(self, command: str): ...

    @abstractmethod
    def read(self) -> str: ...

    @abstractmethod
    def read_all(self) -> str: ...

    def query(self, command: str) -> str:
        self.write(command)
        return self.read()



import time

from WeeLab.connections.base import InstrumentConnection

default_mictosteps = 8
default_steps_per_rotation = 200
default_screw_pitch = 5
default_max_speed = 5
default_acceleration = 2


class Stage:
    tries_to_save = 3

    def __init__(self, conn: InstrumentConnection):
        self.conn = conn
        self.init()

    def __del__(self):
        if self.conn is not None:
            self.save_state()
            self.close()
        else:
            print("No stage to disconnect. Quitting...")

    def init(self):
        self.connected = False
        self.enabled = False
        self.current_position = None
        self.isrunning = (False, False)
        self.speed = (0, 0)
        self.limiters = ((False, False), (False, False))
        # self.connect()
        self.configure()

    def enable(self, en_state):
        self.conn.write(f"en {en_state} ")
        line = self.conn.read()
        if "En changed " not in line:
            raise Exception("Could not enable stage")
        self.enabled = line.decode().split()[3] == "1"

    def configure(self):
        self.steps_per_rotation = default_steps_per_rotation  # steps
        self.screw_pitch = default_screw_pitch  # mm
        self.microsteps = default_mictosteps  # steps

        self.max_speed = default_max_speed  # mm/s
        self.acceleration = default_acceleration  # mm/s^2

        self.steps_per_mm = self.steps_per_rotation * self.microsteps / self.screw_pitch
        self.max_speed_steps = self.max_speed * self.steps_per_mm
        self.acceleration_steps = self.acceleration * self.steps_per_mm

        self.conn.write(f"setvmax {self.max_speed_steps} {self.max_speed_steps} ")
        self.conn.write(f"setacc {self.acceleration_steps} {self.acceleration_steps} ")

        self.conn.write(f"currpos ")
        line = self.conn.read()
        self.current_position = int(line.split()[1]), int(line.split()[2])

    def set_vmax(self, vmax_x):
        if vmax_x is None:
            return
        self.max_speed = vmax_x
        self.max_speed_steps = self.max_speed * self.steps_per_mm
        self.conn.write(f"setvmax {self.max_speed_steps} {self.max_speed_steps} ")

    def set_acc(self, acc_x):
        if acc_x is None:
            return
        self.acceleration = acc_x
        self.acceleration_steps = self.acceleration * self.steps_per_mm
        self.conn.write(f"setacc {self.acceleration_steps} {self.acceleration_steps} ")

    def set_microsteps(self, microsteps):
        if microsteps is None:
            return
        if microsteps not in [
            1,
            2,
            4,
            8,
            16,
            32,
            64,
            128,
            5,
            10,
            20,
            25,
            40,
            50,
            100,
            125,
        ]:
            return
        self.microsteps = microsteps
        self.steps_per_mm = self.steps_per_rotation * self.microsteps / self.screw_pitch
        self.max_speed_steps = self.max_speed * self.steps_per_mm
        self.acceleration_steps = self.acceleration * self.steps_per_mm
        self.conn.write(f"setvmax {self.max_speed_steps} {self.max_speed_steps} ")
        self.conn.write(f"setacc {self.acceleration_steps} {self.acceleration_steps} ")

    def move(self, x, y, mode="mm"):
        self.current_target = (x, y)
        self.this_init_positon = self.current_position
        if mode == "mm":
            x = int(x * self.steps_per_mm)
            y = int(y * self.steps_per_mm)
        elif mode == "steps":
            x = int(x)
            y = int(y)
        self.conn.write(f"move {x} {y} ")

    def move_axis(self, axis, value, mode="mm"):
        self.this_init_positon = self.current_position
        if axis == "x":
            self.current_target = (value, self.current_position[1])
            command = "xmove"
        elif axis == "y":
            self.current_target = (self.current_position[0], value)
            command = "ymove"

        if mode == "mm":
            value = int(value * self.steps_per_mm)
        elif mode == "steps":
            value = int(value)

        self.conn.write(f"{command} {value} ")

    def ask_position(self):
        line = self.conn.query("currpos ")
        # self.current_position = int(line.decode().split()[1]), int(line.decode().split()[2])
        self.current_position = (
            int(line.split()[1]) / self.steps_per_mm,
            int(line.split()[2]) / self.steps_per_mm,
        )
        return self.current_position

    def refresh_status(self):
        self.conn.write(f"status ")
        line = self.conn.read()
        messages = line.split("|")
        # print(line)
        if len(messages) < 5:
            return
        self.enabled = messages[0].split()[1] == "1"
        self.isrunning = (messages[1].split()[1] == "1", messages[1].split()[2] == "1")
        self.current_position = (
            int(messages[2].split()[1]) / self.steps_per_mm,
            int(messages[2].split()[2]) / self.steps_per_mm,
        )
        self.speed = (
            float(messages[3].split()[1]) / self.steps_per_mm,
            float(messages[3].split()[2]) / self.steps_per_mm,
        )
        self.limiters = (
            (messages[4].split()[1] == "1", messages[4].split()[2] == "1"),
            (messages[4].split()[3] == "1", messages[4].split()[4] == "1"),
        )
        # print(self.enabled, self.isrunning, self.current_position, self.speed, self.limiters)

    def print_status(self):
        self.refresh_status()
        print(f"Enabled: {self.enabled}")
        print(f"Is running: {self.isrunning}")
        print(f"Current position: {self.current_position}")
        print(f"Speed: {self.speed}")
        print(f"Limiters: {self.limiters}")

    def set_position(self, axis=None, position=(0, 0)):
        if axis == "x":
            self.conn.write(f"setxcurpos {position*self.steps_per_mm}")
        elif axis == "y":
            self.conn.write(f"setycurpos {position*self.steps_per_mm}")
        else:
            self.conn.write(f"setxcurpos {position[0]*self.steps_per_mm}")
            self.conn.write(f"setycurpos {position[1]*self.steps_per_mm}")
        return self.current_position

    def save_state(self):
        for i in range(self.tries_to_save):
            self.conn.write(f"save ")
            line = self.conn.read()
            if "saved" in line:
                print("State saved successfully")
                return
            else:
                print("Error saving state")
        return

    def stop(self):
        # self.ser.write(f"xstop ".encode())
        # self.ser.write(f"ystop ".encode())
        self.conn.write(f"stop ")

    def close(self):
        self.conn.close()

    def wait_arrived(self):
        self.refresh_status()
        time.sleep(0.5)
        while self.isrunning[0] or self.isrunning[1]:
            self.refresh_status()
            time.sleep(0.5)


# stage = Stage("COM4")



from typing import Dict


class DeviceManager:
    def __init__(self):
        self.devices: Dict[str, object] = {}

    def register(self, name: str, device):
        self.devices[name] = device

    def get(self, name: str):
        return self.devices.get(name)



from abc import ABC, abstractmethod
from typing import Dict
import logging


from WeeLab.const import *
from WeeLab.connections.serial import SerialConnection
from WeeLab.connections.visa import VisaConnection
from WeeLab.connections.vxi11conn import Vxi11Connection
from WeeLab.devices.tek_afg import Tektronix_AFG3252
from WeeLab.devices.caen_a7585 import CAEN_A7585
from WeeLab.devices.lecroy9104 import LeCroyWaveRunner
from WeeLab.devices.pilas import NktPilasLaser
from WeeLab.manager import DeviceManager


# ---------------- Example Setup ----------------
def initialize_devices():
    manager = DeviceManager()

    lecroy_conn = Vxi11Connection(LECROY_IP)
    nkt_conn = SerialConnection(PILAS_COM)
    caen_conn = SerialConnection(CAEN_COM)
    tektronix_conn = VisaConnection(TEKTRONIX_ID)

    manager.register("lecroy", LeCroyWaveRunner(lecroy_conn))
    manager.register("nkt", NktPilasLaser(nkt_conn))
    manager.register("caen", CAEN_A7585(caen_conn))
    manager.register("afg", Tektronix_AFG3252(tektronix_conn))

    return manager


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    devices = initialize_devices()
    lecroy = devices.get("lecroy")
    lecroy.print_id()



from WeeLab.devices.lecroy9104 import LeCroyWaveRunner


def test_lecroy_waveform_read(mock_connection):
    scope = LeCroyWaveRunner(mock_connection)
    scope.find_scale(1)
    scope.getMeasure(1, "MEAN")
    scope.clear_sweeps()
    assert mock_connection.last_command == "vbs 'app.clearsweeps' "



import time
from WeeLab.connections.base import InstrumentConnection

"""
NktPilasLaser Control Module
This module provides an interface to control the Nkt Pilas laser,
allowing for enabling/disabling the laser, setting the tune, trigger mode,
and frequency, as well as retrieving the status of the laser.
Example usage:
```python
from LabPack.devices.pilas import NktPilasLaser
from LabPack.connections.serial import SerialConnection
pilas = NktPilasLaser(
    conn=SerialConnection(port="COM4", baudrate=19200))
pilas.enable()  # Enable the laser
pilas.set_tune(500)  # Set the tune to 500
pilas.set_trigger_mode("internal")  # Set trigger mode to internal
pilas.set_frequency(20_000_000)  # Set frequency to 20 MHz
status = pilas.get_status()  # Get the status of the laser
"""


class NktPilasLaser:
    def __init__(self, conn: InstrumentConnection):
        """
        Initialize the Nkt Pilas laser.
        :param conn: An instance of InstrumentConnection (e.g., SerialConnection).
        Example:
        ``` python
        from LabPack.devices.pilas import NktPilasLaser
        from LabPack.connections.serial import SerialConnection
        pilas = NktPilasLaser(
            conn=SerialConnection(port="COM4", baudrate=19200))
        ```
        """

        self.conn = conn

    def enable(self):
        self.conn.query("ld=1")

    def disable(self):
        self.conn.query("ld=0")

    def set_tune(self, tune: float):
        if not (0 <= tune <= 999):
            raise ValueError("Tune must be between 0 and 999")
        self.conn.query(f"tune={int(tune)}")

    def set_trigger_mode(self, mode: str):
        if mode not in ["internal", "external"]:
            raise ValueError("Mode must be 'internal' or 'external'")
        if mode == "internal":
            self.conn.query(f"ts=0")
        elif mode == "external":
            self.conn.query(f"ts=2")

    def set_frequency(self, frequency: float):
        if not (0 <= frequency <= 40_000_000):
            raise ValueError("Frequency must be between 0 and 40 MHz")
        self.conn.query(f"f={int(frequency)}")

    def get_status(self) -> str:
        self.conn.write("state?")
        time.sleep(1)  # Allow time for the command to process
        return self.conn.read_all().strip()



from WeeLab.connections.base import InstrumentConnection


class SerialConnection(InstrumentConnection):
    def __init__(self, port: str, baudrate: int = 9600):
        import serial

        self.ser = serial.Serial(port, baudrate, timeout=1)

    def write(self, command: str):
        self.ser.write((command + "\n").encode())

    def read(self) -> str:
        return self.ser.readline().decode().strip()

    # def read_all(self) -> str:
    #     return self.ser.read_all().decode()

    def read_all(self) -> str:
        lines = []
        while True:
            try:
                line = self.ser.readline().decode().strip()
            except UnicodeDecodeError:
                line = self.ser.readline().decode().strip()
            # line = self.ser.readline().decode().strip()
            if not line:
                break
            lines.append(line)
        return "\n".join(lines)

    def close(self):
        self.ser.close()

    def get_serial_number(self) -> str:
        return self.ser.serial_number if hasattr(self.ser, "serial_number") else None

    def __del__(self):
        self.close()



from WeeLab.connections.base import InstrumentConnection


class Vxi11Connection(InstrumentConnection):
    def __init__(self, ip: str):
        import vxi11

        # self.inst = self.rm.open_resource(resource_name)
        self.inst = vxi11.Instrument(ip)

    def write(self, command: str):
        self.inst.write(command)

    def read(self) -> str:
        return self.inst.read()

    def read_all(self) -> str:
        return self.inst.read_all()

    def query(self, command: str) -> str:
        return self.inst.ask(command)



from WeeLab.devices.caen_a7585 import CAEN_A7585


def test_caen_voltage_command(mock_connection):
    caen = CAEN_A7585(mock_connection)
    caen.set_voltage(80)
    # caen.set_voltage_modulating_ramp(23, 20)
    assert mock_connection.last_command == "AT+SET,0,1"



from WeeLab.devices.tek_afg import Tektronix_AFG3252


def test_afg_voltage_command(mock_connection):
    afg = Tektronix_AFG3252(mock_connection)
    afg.do_set_voltage_ch1(5)
    afg.set_frequency_ch1(10)
    # caen.set_voltage_modulating_ramp(23, 20)
    assert mock_connection.last_command == "source1:frequency 10.000000"



import time
from WeeLab.connections.base import InstrumentConnection
from WeeLab.interfaces.oscilloscope import Oscilloscope

import os


class LeCroyWaveRunner(Oscilloscope):
    def __init__(self, conn: InstrumentConnection):
        self.conn = conn

    def clear_sweeps(self):
        self.conn.write(r"""vbs 'app.clearsweeps' """)

    def save_insitu_waveforms(self):
        self.conn.query(r"""vbs 'app.SaveRecall.Waveform.SaveFile' """)

    def get_waveforms(self, output_folder, filename):
        self.conn.write("TRMD STOP")

        # method 2 in 1 step
        deltaT_suffix = "F1"
        LG_sh_amplitude_master_suffix = "F3"
        LG_sh_amplitude_slave_suffix = "F2"
        TOT_width_master_suffix = "F4"
        TOT_width_slave_suffix = "F5"
        file_start_index = 12
        file_stop_index = -1
        index_step = 14
        trace_list = [
            deltaT_suffix,
            LG_sh_amplitude_master_suffix,
            LG_sh_amplitude_slave_suffix,
            TOT_width_master_suffix,
            TOT_width_slave_suffix,
        ]

        trace_data_list = []
        for trace in trace_list:
            trace_data_list.append(
                self.conn.query(trace + ":INSP? 'DATA_ARRAY_2',FLOAT").replace("\r\n", "")
            )  # truncate header and footer

            # trace_data_list.append(scopeLecroy.oscillo.query(trace + ":INSP? 'DATA_ARRAY_2',FLOAT")
            #             .replace("\r\n", '')[file_start_index:file_stop_index]  # truncate header and footer
            #             .replace('   ', '\n').replace('  ', '\n'))  # deal with spaces for + and - float

        events_number = int((len(trace_data_list[0]) - 10) / 14)
        # all col have the same events_number since trend mode = average and scope stopped between data fetches "

        # csv_file = ','.join([f'{trace=}'.split('=')[0] for trace in trace_list])
        csv_file = "deltaT,LG_sh_amplitude_master,LG_sh_amplitude_slave,TOT_width_master,TOT_width_slave\n"
        for i in range(events_number):
            for ti, trace in enumerate(trace_data_list):
                csv_file += trace[12 + i * 14 : 10 + 13 + i * 14]
                if ti != len(trace_data_list) - 1:
                    csv_file += ","
                else:
                    csv_file += "\n"

        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        with open(output_folder + filename, "w") as text_file:
            text_file.write(csv_file)

        text_file.close()

        pass
        self.conn.write("TRMD NORM")

        # method 1 in 2 steps
        # scopeLecroy.oscillo.query("STO ALL_DISPLAYED,FILE")
        # for trace in trace_list:
        #     with open(trace+output_file, "w") as text_file:
        # scopeLecroy.oscillo.query("TRFL? DISK,HDD,FILE,'D:\\file.txt'")

        # method 3
        # to get byte to save and recall, not to read
        # scopeLecroy.oscillo.ask_raw(str("CHDR SHORT; C1:WaveForm?").encode("utf-8"))

    def getMeasure(self, channel: int, measureType: str, exponent=0):
        """
        Get the measure value from the oscilloscope.
        :param channel: Channel number (1, 2, 3, or 4)
        :param measureType: Type of measure to retrieve (e.g., "mean", "sdev", "num")
        :param exponent: Exponent to apply to the measure value (default is 0)

        :return: The measure value as a float
        """
        validMeasureType = {"mean", "sdev", "num", "min", "max"}
        measureType = measureType.lower()
        if measureType not in validMeasureType:
            raise ValueError(
                "Lecroy measure function call has received a wrong measure type :"
                + str(measureType)
            )
        stringAnswer = self.conn.query(
            f"""vbs? 'return=app.measure.p{channel}.{measureType}.result.value' """
        )
        try:
            # val = float(stringAnswer.split()[1]) * 10 ** exponent
            val = float(stringAnswer.split()[1])
        except:
            val = 0

        # val = '%.3f' % val
        return val

    def getSetupChannel(self, channel):
        return self.conn.query("C" + str(channel) + ":PAVA?")

    def getSetupMeasure(self, measureNumber):
        return self.conn.query("PACU? " + str(measureNumber))

    def setTimebase(self, time):  # unit : second
        self.conn.write(
            r"""vbs? 'app.Acquisition.Horizontal.HorScale = {:.0e}' """.format(time)
        )

    def write(self, measure):
        self.conn.write(measure)

    def query(self, measure):
        return self.conn.query(measure)

    def read(self):
        return self.conn.read()

    def save_setup(self, filepath):
        self.conn.write("*SAV 2")
        setup = self.conn.query("PANEL_SETUP?")
        output_folder = os.path.dirname(filepath)
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        with open(filepath, "w") as text_file:
            text_file.write(setup)
        text_file.close()

    def load_setup(self, filepath):
        with open(filepath, "r") as text_file:
            read_setup = text_file.read()
        text_file.close()
        self.conn.write(read_setup)
        self.conn.write("*RCL 2")

    def query_float(self, command):
        """
        Query the oscilloscope for a float value using a VBS command.
        :param command: The VBS command to execute
        :return: The float value read from the oscilloscope
        """
        self.write(command)
        try:
            return float(self.read().split()[-1])
        except:
            time.sleep(1)
            return self.query_float(command)

    def find_scale(self, channel):
        """
        Find the vertical scale for a given channel.
        :param channel: Channel number (1, 2, 3, or 4)
        :return: The scale value as a float
        """
        self.write(f"VBS app.Acquisition.C{channel}.FindScale")

    def find_scale_man(self, pos=False, channel=1, pkpk_channel=2):
        """
        Find the vertical scale and offset for a given channel based on the peak-to-peak amplitude.
        :param pos: If True, assumes a positive signal, otherwise assumes a negative signal (default is False)
        :param channel: Channel number to set the scale and offset for (default is 1)
        :param pkpk_channel: Channel number to measure the peak-to-peak amplitude from (default is 2)
        """
        self.find_scale(channel=channel)
        self.clear_sweeps()
        time.sleep(3)
        ampl = self.query_float(
            f"VBS? 'return=app.Measure.P{pkpk_channel}.Mean.Result.Value"
        )
        if pos:
            sign = -1
        else:
            sign = 1

        if ampl < 0.01:
            scale = abs(ampl) / 2
            self.write(f"VBS app.Acquisition.C{channel}.VerScale = {scale}")
            scale = self.query_float(f"VBS? 'return=app.Acquisition.C{channel}.VerScale")
            offset = sign * 2 * scale
            self.write(f"VBS app.Acquisition.C{channel}.VerOffset = {offset}")
            # print(f"Signal amplitude {ampl}: set scale to {scale} and offset to {offset}")
        else:
            scale = abs(ampl) / 4
            self.write(f"VBS app.Acquisition.C{channel}.VerScale = {scale}")
            scale = self.query_float(f"VBS? 'return=app.Acquisition.C{channel}.VerScale")
            offset = sign * 3 * scale
            self.write(f"VBS app.Acquisition.C{channel}.VerOffset = {offset}")
            # print(f"Signal amplitude {ampl}: set scale to {scale} and offset to {offset}")

    def print_id(self):
        """
        Print the ID of the oscilloscope.
        """
        idn = self.conn.query("*IDN?")
        print(idn)



from os import name
from WeeLab.connections.base import InstrumentConnection
from WeeLab.interfaces.wf_generator import WaveformGenerator

import logging


class Tektronix_AFG3252(WaveformGenerator):
    def __init__(self, conn: InstrumentConnection):
        self.conn = conn
        self.maxpoint = 131072
        self.maxrate = 2e9
        self.max_offset = 2.5
        self.min_offset = -2.5
        self.max_amplitude = 5.0
        self.min_amplitude = 0.05

        logging.debug(__name__ + " :  Get state of the channel 1 output on/off")

    def get_all(self):
        """
        Get all parameters of the intrument

        Input:
            None

        Output:
            None
        """
        logging.info(__name__ + " : get all")
        self.do_get_period_ch1()
        self.do_get_amplitude_ch1()
        self.do_get_offset_ch1()
        self.do_get_phase_ch1()
        self.do_get_status_ch1()
        self.do_get_ncycles_ch1()
        self.do_get_interval()
        self.do_get_reference()

    #########################################################
    #
    #
    #                       Interval between each pulse of the burst mode
    #
    #
    #########################################################

    def do_get_interval(self):
        """
        Get the interval between each cycle

        Input:
            None

        Output:
            interval (float) : The interval between each cycle
        """
        logging.debug(__name__ + " : Get the interval between each cycle")
        return float(self.conn.query("trigger:timer?"))

    def do_set_interval(self, interval=1.0):
        """
        Set the interval between each cycle

        Input:
            interval (float) : interval between each cycle [s]

        Output:
            None
        """
        logging.debug(
            __name__ + " : Set the interval between each cycle to %.6f" % (interval)
        )
        self.conn.write("trigger:timer " + str(interval) + "s")

    #########################################################
    #
    #
    #                       Number of cycle per period
    #
    #
    #########################################################

    def do_get_ncycles_ch1(self):
        """
        Get the number of cycle per period for the channel 1

        Input:
            None

        Output:
            ncycle (Int) : The number of cycle per period of the channel 1
        """
        logging.debug(
            __name__ + " : Get the number of cycle per period for the channel  1"
        )
        return float(self.conn.query("source1:burst:ncycles?"))

    def do_set_ncycles_ch1(self, ncycles=0):
        """
        Set the number of cycle for the channel 1

        Input:
            ncycle (Int) : Number if cycle per period

        Output:
            None
        """
        logging.debug(
            __name__
            + " : Set the number of cycle per period for channel  1 to %.6f" % (ncycles)
        )
        self.conn.write("source1:burst:ncycles " + str(int(ncycles)))

    #########################################################
    #
    #
    #                       Offset
    #
    #
    #########################################################

    def do_get_offset_ch1(self):
        """
        Get the offset for the channel 1

        Input:
            None

        Output:
            Offset (Float) : The offset of the channel 1 in Volts
        """
        logging.debug(__name__ + " : Get offset of channel  1")
        return float(self.conn.query("SOURCE1:VOLTAGE:LEVEL:IMMEDIATE:OFFSET?"))

    def do_set_offset_ch1(self, offset=0):
        """
        Set the offset for the channel 1 to a defined value

        Input:
            offset (float) : offset in Volts

        Output:
            None
        """
        logging.debug(__name__ + " : Set offset of channel  1 to %.6f" % (offset))
        self.conn.write("SOURCE1:VOLTAGE:LEVEL:IMMEDIATE:OFFSET %.6f" % (offset))

    #########################################################
    #
    #
    #                   Amplitude
    #
    #
    #########################################################

    def do_get_amplitude_ch1(self):
        """
        Get the amplitude for the channel 1

        Input:
            None

        Output:
            amplitude (Float) : The amplitude of the channel 1 in Volts, peak to peak
        """

        logging.debug(__name__ + " : Get amplitude of channel  1")

        return float(self.conn.query("source1:voltage:level:immediate:amplitude?"))

    def do_set_amplitude_ch1(self, amplitude=0.1):
        """
        Set the amplitude for the channel 1 to a defined value

        Input:
            amplitude (float) : amplitude in Volts, peak to peak

        Output:
            None
        """
        logging.debug(__name__ + " : Set amplitude of channel  1 to %.6f" % (amplitude))
        self.conn.write("source1:voltage:level:immediate:amplitude %.6f" % (amplitude))

    #########################################################
    #
    #
    #                   Phase
    #
    #
    #########################################################

    def do_get_phase_ch1(self):
        """
        Get the phase for the channel 1

        Input:
            None

        Output:
            phase (Float) : The phase of the channel 1 [rad]
        """
        logging.debug(__name__ + " : Get phase of channel  1")
        return float(self.conn.query("source1:phase?"))

    def do_set_phase_ch1(self, phase=0.0):
        """
        Set the phase for the channel 1 to a defined value

        Input:
            phase (float) : phase of the channel 1 [rad]

        Output:
            None
        """
        logging.debug(__name__ + " : Set phase of channel  1 to %.6f" % (phase))
        self.conn.write("source1:phase:adjust %.6f" % (phase))

    #########################################################
    #
    #
    #                   Frequency
    #
    #
    #########################################################

    def get_frequency_ch1(self):
        """
        Get the frequency for the channel 1

        Input:
            None

        Output:
            frequency (Float) : The frequency of the channel 1 in Hertz
        """
        logging.debug(__name__ + " : Get frequency of channel 1")
        return float(self.conn.query("source1:frequency?"))

    def set_frequency_ch1(self, frequency=1e6):
        """
        Set the frequency for the channel 1 to a defined value

        Input:
            frequency (float) : frequency in Hertz

        Output:
            None
        """
        logging.debug(
            __name__ + " : Set the frequency of channel  1 to %.6f" % (frequency)
        )
        self.conn.write("source1:frequency %.6f" % (frequency))

    #########################################################
    #
    #
    #                   Period
    #
    #
    #########################################################

    def do_get_period_ch1(self):
        """
        Get the period for the channel 1

        Input:
            None

        Output:
            period (Float) : The frequency of the channel 1 in Hertz
        """
        logging.debug(__name__ + " : Get period of channel 1")
        return float(self.conn.query("source1:pulse:period?"))

    def do_set_period_ch1(self, period=1):
        """
        Set the period for the channel 1 to a defined value

        Input:
            frequency (float) : period in Hertz

        Output:
            None
        """
        logging.debug(__name__ + " : Set the period of channel  1 to %.6f" % (period))
        self.conn.write("source1:pulse:period %.6f" % (period))

    #########################################################
    #
    #
    #                   Burst mode
    #
    #
    #########################################################

    def set_run_mode_burst_ch1(self):
        """
        Set the run mode for the channel 1 to burst mode

        Input:
            None

        Output:
            None
        """
        logging.debug(__name__ + " : Set the run mode for the channel 1 to burst mode")
        self.conn.write("source1:burst:state on")

    def set_burst_mode_triger_ch1(self):
        """
        Set the burst mode for the channel 1 to triger mode

        Input:
            None

        Output:
            None
        """
        logging.debug(__name__ + " : Set the burst mode for the channel 1 to triger mode")
        self.conn.write("source1:burst:mode trigered")

    def set_burst_mode_ncycle_ch1(self, number_cycle=1):
        """
        Set the number of cycle of the burst mode for the channel 1

        Input:
            number of cycle (int) : Number of cycle in burst mode
                - 0  : set the number of cycle to infinite count
                - !0 : set the number of cycle to defined value

        Output:
            None
        """

        if int(number_cycle) == 0:

            logging.debug(
                __name__
                + " : Set the number of cycle of the burst mode for the channel 1"
            )
            self.conn.write("source1:burst:ncycles infinity")
        else:

            logging.debug(
                __name__
                + " : Set the number of cycle of the burst mode for the channel 1"
            )
            self.conn.write("source1:burst:ncycles  %.6f" % (int(number_cycle)))

    #########################################################
    #
    #
    #                   Trigger
    #
    #
    #########################################################

    def set_triger_source_external(self):
        """
        Set the trigger to an external source

        Input:
            None

        Output:
            None
        """
        logging.debug(__name__ + " : Set the trigger to an external source")
        self.conn.write("trigger:sequence:source external")

    def set_triger_source_internal(self):
        """
        Set the trigger to an internal source

        Input:
            None

        Output:
            None
        """
        logging.debug(__name__ + " : Set the trigger to an internal source")
        self.conn.write("trigger:sequence:source timer")

    def send_triger_event(self):
        """
        Send a trigger event

        Input:
            None

        Output:
            None
        """
        logging.debug(__name__ + " : Send a trigger event")
        self.conn.write("trigger:sequence:immediate")

    #########################################################
    #
    #
    #                   Memory
    #
    #
    #########################################################

    def set_transfert_ememory_user1(self):
        """
        Set the transfert of the waveform present in the ememory to the user 1 to save it

        Input:
            None

        Output:
            None
        """
        logging.debug(
            __name__ + " : Set the transfert of the waveform present in the ememory"
        )
        self.conn.write("trace:copy user1,ememory")

    def set_function_user1_ch1(self):
        """
        Set the function of the channel 1 to user1

        Input:
            None

        Output:
            None
        """
        logging.debug(__name__ + " : Set the function of the channel 1 to user1")
        self.conn.write("source1:function user1")

    def set_arbitrary_waveform_memory_ch1(self):
        """
        Set the device to the arbitrary waveform menu for the channel 1

        Input:
            None

        Output:
            None
        """
        logging.debug(
            __name__
            + " : Set the device to the arbitrary waveform menu for the channel 1"
        )
        self.conn.write("source1:function ememory")

    ##################################################################
    #
    #
    #                   Status
    #
    #
    ##################################################################

    def do_set_status_ch1(self, state="off"):
        """
        Set channel 1 output on/off. By default off.

        Input:
            state (string) : switch on/off the channel

        Output:
            None
        """
        logging.debug(__name__ + " :  Set channel 1 output on/off")

        if state.lower() == "on":
            self.conn.write("output1:state 1")
        elif state.lower() == "off":
            self.conn.write("output1:state 0")
        else:
            logging.debug(__name__ + " : Status can only set on or off")
            return "Error : Status can only set on or off"

    def do_get_status_ch1(self):
        """
        Get the state of the channel 1 output on/off

        Input:
            state (string) : switch on/off the channel

        Output:
            None
        """
        logging.debug(__name__ + " :  Get state of the channel 1 output on/off")

        if self.conn.query("output1:state?") == "1":
            return "on"
        else:
            return "off"

    ##################################################################
    #
    #
    #                   reference
    #
    #
    ##################################################################

    def do_set_reference(self, val="external"):
        """
        Set the origin of the reference

        Input:
            val (string) : switch the clock on the external or internal reference

        Output:
            None
        """
        logging.debug(__name__ + " :  Set the clock")

        if val.lower() == "external" or val.lower() == "internal":
            self.conn.write("source:roscillator:source " + str(val))
        else:
            raise ValueError('The reference has to be either "Internal" or "external"')

    def do_get_reference(self):
        """
        Get the reference

        Input:
            None

        Output:
            state (string) : reference
        """
        logging.debug(__name__ + " :  Get the reference")
        return self.conn.query("source:roscillator:source?")
        # if self.conn.query('source:roscillator:source?') == 'EXT':
        #     return 'External'
        # else:
        #     return 'Internal'

    ###################################################################
    #
    #
    #                   Waveform
    #
    #
    ####################################################################

    def set_waveform_ch1(self, wave):
        """
        Sends a complete waveform.

        Input:
            w (float[numpoints]) : waveform in V

        Output:
            None
        """
        logging.debug(__name__ + " : set waveform")

        # We check if the waveform if no to long
        if len(wave) > self.maxpoint or len(wave) < 2:

            logging.debug(
                __name__ + " :The waveform is too long or too short (number of points)."
            )
            return "Errors : The waveform is too long or too short (number of points)."

        # We check if the waveform is feasible
        if (
            max(wave) > self.max_amplitude / 2 + self.max_offset
            or min(wave) < -self.max_amplitude / 2 - self.max_offset
            or max(wave) - min(wave) > self.max_amplitude
        ):

            logging.debug(__name__ + " : The waveform is out of the range.")
            return "Errors : The waveform is out of the range."

        # We calculate the amplitude and the offset of the tektro
        amplitude = max(wave) - min(wave)
        offset = (max(wave) + min(wave)) / 2

        # We tune these values
        self.set_amplitude_ch1(amplitude)
        self.set_offset_ch1(offset)

        # We transform all points of the waveforme which are in V in bytes
        wave_bytes = ""
        for i in wave:

            # First we tranform V into bins
            j = int(((i + abs(min(wave))) * (2**14 - 1)) / amplitude)

            wave_bytes = wave_bytes + struct.pack(">H", j)

        # We research the beginning of the message for the tektro
        number_bytes = int(len(wave)) * 2
        number_digits = len(str(number_bytes))

        # We create the final message for the device
        mes = "trace ememory,#" + str(number_digits) + str(number_bytes) + wave_bytes

        # we prepare the device to be configured
        self.set_arbitrary_waveform_memory_ch1()
        self.conn.write(mes)

    def do_set_function_ch1(self, function="sin"):
        """
        Set the function of the channel 1

        Input:
            function (string) : function of the channel 1

        Output:
            None
        """
        logging.debug(__name__ + " : Set the function of the channel 1")
        self.conn.write("source1:function " + str(function))

    def do_set_width_ch1(self, width=0.5):
        """
        Set the duty cycle of the pulse

        Input:
            duty_cycle (float) : duty cycle in seconds

        Output:
            None
        """
        logging.debug(__name__ + " : Set the duty cycle of the pulse")
        self.conn.write(f"SOURce1:PULSe:hold width")
        self.conn.write(f"SOURce1:PULSe:width {width}ns")

    def do_set_high_level(self, level=0.5):
        """
        Set the high level of the pulse

        Input:
            level (float) : high level in seconds

        Output:
            None
        """
        logging.debug(__name__ + " : Set the high level of the pulse")
        assert level >= 0, "Level must be non-negative"
        assert level <= 5, "Level must be less than or equal to 5"
        self.conn.write(f"SOURce1:voltage:level:HIGH {level}")

    def do_set_level_low(self, level=0.5):
        """
        Set the low level of the pulse

        Input:
            level (float) : low level in seconds

        Output:
            None
        """
        logging.debug(__name__ + " : Set the low level of the pulse")
        assert level >= 0, "Level must be non-negative"
        assert level <= 5, "Level must be less than or equal to 5"
        self.conn.write(f"SOURce1:voltage:level:LOW {level}")

    def do_set_voltage_ch1(self, voltage=0.5):
        """
        Set the voltage of the channel 1

        Input:
            voltage (float) : voltage of the channel 1

        Output:
            None
        """
        logging.debug(__name__ + " : Set the voltage of the channel 1")
        self.conn.write(f"SOURce1:voltage:level:IMMediate:HIGH {voltage}")
        self.conn.write(f"SOURce1:voltage:level:IMMediate:LOW {0}")

    def do_set_delay_ch1(self, delay=20):
        """
        Set the delay of the pulse

        Input:
            delay (float) : delay in seconds

        Output:
            None
        """
        logging.debug(__name__ + " : Set the delay of the pulse")
        self.conn.write(f"SOURce1:PULSe:DELay {delay}ns")


import pyvisa
import types
import logging
from numpy import pi
import struct
import time


# class Instrument:
#     def __init__(self, name, tags=None):
#         self.name = name
#         self.tags = tags if tags is not None else []
#         self.conn = None
#         self.parameters = {}
#         pass

#     def add_parameter(self, name, flags, units, minval, maxval, type):
#         self.parameters[name] = {
#             "flags": flags,
#             "units": units,
#             "minval": minval,
#             "maxval": maxval,
#             "type": type,
#         }

#     def get_all(self):
#         """
#         Get all parameters of the intrument

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.info(__name__ + " : get all")
#         self.do_get_period_ch1()
#         self.do_get_amplitude_ch1()
#         self.do_get_offset_ch1()
#         self.do_get_phase_ch1()
#         self.do_get_status_ch1()
#         self.do_get_ncycles_ch1()
#         self.do_get_interval()
#         self.do_get_reference()


# class Tektronix_AFG3252_old(Instrument):
#     """
#     This is the python driver for the Tektronix AFG3252
#     Arbitrary Waveform Generator

#     Usage:
#     Initialize with
#     <name> = instruments.create('name', 'Tektronix_AFG3252', address='<GPIB address>')
#     """

#     def __init__(self, name, address, reset=False):  # , clock=1e9, numpoints=1000):
#         """
#         Initializes the AFG3252.

#         Input:
#             name (string)    : name of the instrument
#             address (string) : GPIB address
#             reset (bool)     : resets to default values, default=false

#         Output:
#             None
#         """

#         logging.debug(__name__ + " : Initializing instrument")
#         Instrument.__init__(self, name, tags=["physical"])
#         rm = pyvisa.ResourceManager()

#         self._address = address
#         try:
#             self._visainstrument = rm.open_resource(self._address)
#         except:
#             raise SystemExit

#         # self.add_parameter('period_ch1', flags=Instrument.FLAG_GETSET, units='s', minval=16.67e-9, maxval=1e3, type=types.FloatType)
#         # self.add_parameter('amplitude_ch1', flags=Instrument.FLAG_GETSET, units='V', minval=0.05, maxval=5.0, type=types.FloatType)
#         # self.add_parameter('offset_ch1', flags=Instrument.FLAG_GETSET, units='V', minval=-2.5, maxval=2.5, type=types.FloatType)
#         # self.add_parameter('phase_ch1', flags=Instrument.FLAG_GETSET, units='rad', minval=-pi, maxval=pi, type=types.FloatType)
#         # self.add_parameter('status_ch1', option_list=('off', 'on'), flags=Instrument.FLAG_GETSET, type=types.StringType)
#         # self.add_parameter('ncycles_ch1', minval=1, maxval= 1000000, flags=Instrument.FLAG_GETSET, type=types.IntType)
#         # self.add_parameter('interval', minval=1e-3, maxval= 500, units='s', flags=Instrument.FLAG_GETSET, type=types.FloatType)
#         # self.add_parameter('reference', option_list=['Internal', 'External'], flags=Instrument.FLAG_GETSET, type=types.StringType)

#         #        self.add_function('get_all')
#         # self.add_function('reset')

#         self.maxpoint = 131072
#         self.maxrate = 2e9

#         self.max_offset = 2.5
#         self.min_offset = -2.5

#         self.max_amplitude = 5.0
#         self.min_amplitude = 0.05

#         if reset:

#             self.reset()

#         self.get_all()

#     # Functions
#     def reset(self):
#         """
#         Resets the instrument to default values

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.info(__name__ + " : Reset the instrument")
#         self._visainstrument.write("*RST")

#     def get_all(self):
#         """
#         Get all parameters of the intrument

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.info(__name__ + " : get all")
#         self.do_get_period_ch1()
#         self.do_get_amplitude_ch1()
#         self.do_get_offset_ch1()
#         self.do_get_phase_ch1()
#         self.do_get_status_ch1()
#         self.do_get_ncycles_ch1()
#         self.do_get_interval()
#         self.do_get_reference()

#     def get_maxpoint(self):
#         """
#         Return the max number of point that the tektro can handle

#         Input:
#             None

#         Output:
#             maxpoint (int) : the max number of point that the tektro can handle
#         """

#         return self.maxpoint

#     def get_maxrate(self):
#         """
#         Return the maw rate of the tektro

#         Input:
#             None

#         Output:
#             maxrate (int) : the max rate of the tektro
#         """

#         return self.maxrate

#     #########################################################
#     #
#     #
#     #                       Interval between each pulse of the burst mode
#     #
#     #
#     #########################################################

#     def do_get_interval(self):
#         """
#         Get the interval between each cycle

#         Input:
#             None

#         Output:
#             interval (float) : The interval between each cycle
#         """
#         logging.debug(__name__ + " : Get the interval between each cycle")
#         return float(self._visainstrument.query("trigger:timer?"))

#     def do_set_interval(self, interval=1.0):
#         """
#         Set the interval between each cycle

#         Input:
#             interval (float) : interval between each cycle [s]

#         Output:
#             None
#         """
#         logging.debug(
#             __name__ + " : Set the interval between each cycle to %.6f" % (interval)
#         )
#         self._visainstrument.write("trigger:timer " + str(interval) + "s")

#     #########################################################
#     #
#     #
#     #                       Number of cycle per period
#     #
#     #
#     #########################################################

#     def do_get_ncycles_ch1(self):
#         """
#         Get the number of cycle per period for the channel 1

#         Input:
#             None

#         Output:
#             ncycle (Int) : The number of cycle per period of the channel 1
#         """
#         logging.debug(
#             __name__ + " : Get the number of cycle per period for the channel  1"
#         )
#         return float(self._visainstrument.query("source1:burst:ncycles?"))

#     def do_set_ncycles_ch1(self, ncycles=0):
#         """
#         Set the number of cycle for the channel 1

#         Input:
#             ncycle (Int) : Number if cycle per period

#         Output:
#             None
#         """
#         logging.debug(
#             __name__
#             + " : Set the number of cycle per period for channel  1 to %.6f" % (ncycles)
#         )
#         self._visainstrument.write("source1:burst:ncycles " + str(int(ncycles)))

#     #########################################################
#     #
#     #
#     #                       Offset
#     #
#     #
#     #########################################################

#     def do_get_offset_ch1(self):
#         """
#         Get the offset for the channel 1

#         Input:
#             None

#         Output:
#             Offset (Float) : The offset of the channel 1 in Volts
#         """
#         logging.debug(__name__ + " : Get offset of channel  1")
#         return float(
#             self._visainstrument.query("SOURCE1:VOLTAGE:LEVEL:IMMEDIATE:OFFSET?")
#         )

#     def do_set_offset_ch1(self, offset=0):
#         """
#         Set the offset for the channel 1 to a defined value

#         Input:
#             offset (float) : offset in Volts

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set offset of channel  1 to %.6f" % (offset))
#         self._visainstrument.write(
#             "SOURCE1:VOLTAGE:LEVEL:IMMEDIATE:OFFSET %.6f" % (offset)
#         )

#     #########################################################
#     #
#     #
#     #                   Amplitude
#     #
#     #
#     #########################################################

#     def do_get_amplitude_ch1(self):
#         """
#         Get the amplitude for the channel 1

#         Input:
#             None

#         Output:
#             amplitude (Float) : The amplitude of the channel 1 in Volts, peak to peak
#         """

#         logging.debug(__name__ + " : Get amplitude of channel  1")

#         return float(
#             self._visainstrument.query("source1:voltage:level:immediate:amplitude?")
#         )

#     def do_set_amplitude_ch1(self, amplitude=0.1):
#         """
#         Set the amplitude for the channel 1 to a defined value

#         Input:
#             amplitude (float) : amplitude in Volts, peak to peak

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set amplitude of channel  1 to %.6f" % (amplitude))
#         self._visainstrument.write(
#             "source1:voltage:level:immediate:amplitude %.6f" % (amplitude)
#         )

#     #########################################################
#     #
#     #
#     #                   Phase
#     #
#     #
#     #########################################################

#     def do_get_phase_ch1(self):
#         """
#         Get the phase for the channel 1

#         Input:
#             None

#         Output:
#             phase (Float) : The phase of the channel 1 [rad]
#         """
#         logging.debug(__name__ + " : Get phase of channel  1")
#         return float(self._visainstrument.query("source1:phase?"))

#     def do_set_phase_ch1(self, phase=0.0):
#         """
#         Set the phase for the channel 1 to a defined value

#         Input:
#             phase (float) : phase of the channel 1 [rad]

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set phase of channel  1 to %.6f" % (phase))
#         self._visainstrument.write("source1:phase:adjust %.6f" % (phase))

#     #########################################################
#     #
#     #
#     #                   Frequency
#     #
#     #
#     #########################################################

#     def get_frequency_ch1(self):
#         """
#         Get the frequency for the channel 1

#         Input:
#             None

#         Output:
#             frequency (Float) : The frequency of the channel 1 in Hertz
#         """
#         logging.debug(__name__ + " : Get frequency of channel 1")
#         return float(self._visainstrument.query("source1:frequency?"))

#     def set_frequency_ch1(self, frequency=1e6):
#         """
#         Set the frequency for the channel 1 to a defined value

#         Input:
#             frequency (float) : frequency in Hertz

#         Output:
#             None
#         """
#         logging.debug(
#             __name__ + " : Set the frequency of channel  1 to %.6f" % (frequency)
#         )
#         self._visainstrument.write("source1:frequency %.6f" % (frequency))

#     #########################################################
#     #
#     #
#     #                   Period
#     #
#     #
#     #########################################################

#     def do_get_period_ch1(self):
#         """
#         Get the period for the channel 1

#         Input:
#             None

#         Output:
#             period (Float) : The frequency of the channel 1 in Hertz
#         """
#         logging.debug(__name__ + " : Get period of channel 1")
#         return float(self._visainstrument.query("source1:pulse:period?"))

#     def do_set_period_ch1(self, period=1):
#         """
#         Set the period for the channel 1 to a defined value

#         Input:
#             frequency (float) : period in Hertz

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the period of channel  1 to %.6f" % (period))
#         self._visainstrument.write("source1:pulse:period %.6f" % (period))

#     #########################################################
#     #
#     #
#     #                   Burst mode
#     #
#     #
#     #########################################################

#     def set_run_mode_burst_ch1(self):
#         """
#         Set the run mode for the channel 1 to burst mode

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the run mode for the channel 1 to burst mode")
#         self._visainstrument.write("source1:burst:state on")

#     def set_burst_mode_triger_ch1(self):
#         """
#         Set the burst mode for the channel 1 to triger mode

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the burst mode for the channel 1 to triger mode")
#         self._visainstrument.write("source1:burst:mode trigered")

#     def set_burst_mode_ncycle_ch1(self, number_cycle=1):
#         """
#         Set the number of cycle of the burst mode for the channel 1

#         Input:
#             number of cycle (int) : Number of cycle in burst mode
#                 - 0  : set the number of cycle to infinite count
#                 - !0 : set the number of cycle to defined value

#         Output:
#             None
#         """

#         if int(number_cycle) == 0:

#             logging.debug(
#                 __name__
#                 + " : Set the number of cycle of the burst mode for the channel 1"
#             )
#             self._visainstrument.write("source1:burst:ncycles infinity")
#         else:

#             logging.debug(
#                 __name__
#                 + " : Set the number of cycle of the burst mode for the channel 1"
#             )
#             self._visainstrument.write(
#                 "source1:burst:ncycles  %.6f" % (int(number_cycle))
#             )

#     #########################################################
#     #
#     #
#     #                   Trigger
#     #
#     #
#     #########################################################

#     def set_triger_source_external(self):
#         """
#         Set the trigger to an external source

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the trigger to an external source")
#         self._visainstrument.write("trigger:sequence:source external")

#     def set_triger_source_internal(self):
#         """
#         Set the trigger to an internal source

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the trigger to an internal source")
#         self._visainstrument.write("trigger:sequence:source timer")

#     def send_triger_event(self):
#         """
#         Send a trigger event

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Send a trigger event")
#         self._visainstrument.write("trigger:sequence:immediate")

#     #########################################################
#     #
#     #
#     #                   Memory
#     #
#     #
#     #########################################################

#     def set_transfert_ememory_user1(self):
#         """
#         Set the transfert of the waveform present in the ememory to the user 1 to save it

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.debug(
#             __name__ + " : Set the transfert of the waveform present in the ememory"
#         )
#         self._visainstrument.write("trace:copy user1,ememory")

#     def set_function_user1_ch1(self):
#         """
#         Set the function of the channel 1 to user1

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the function of the channel 1 to user1")
#         self._visainstrument.write("source1:function user1")

#     def set_arbitrary_waveform_memory_ch1(self):
#         """
#         Set the device to the arbitrary waveform menu for the channel 1

#         Input:
#             None

#         Output:
#             None
#         """
#         logging.debug(
#             __name__
#             + " : Set the device to the arbitrary waveform menu for the channel 1"
#         )
#         self._visainstrument.write("source1:function ememory")

#     ##################################################################
#     #
#     #
#     #                   Status
#     #
#     #
#     ##################################################################

#     def do_set_status_ch1(self, state="off"):
#         """
#         Set channel 1 output on/off. By default off.

#         Input:
#             state (string) : switch on/off the channel

#         Output:
#             None
#         """
#         logging.debug(__name__ + " :  Set channel 1 output on/off")

#         if state.lower() == "on":
#             self._visainstrument.write("output1:state 1")
#         elif state.lower() == "off":
#             self._visainstrument.write("output1:state 0")
#         else:
#             logging.debug(__name__ + " : Status can only set on or off")
#             return "Error : Status can only set on or off"

#     def do_get_status_ch1(self):
#         """
#         Get the state of the channel 1 output on/off

#         Input:
#             state (string) : switch on/off the channel

#         Output:
#             None
#         """
#         logging.debug(__name__ + " :  Get state of the channel 1 output on/off")

#         if self._visainstrument.query("output1:state?") == "1":
#             return "on"
#         else:
#             return "off"

#     ##################################################################
#     #
#     #
#     #                   reference
#     #
#     #
#     ##################################################################

#     def do_set_reference(self, val="external"):
#         """
#         Set the origin of the reference

#         Input:
#             val (string) : switch the clock on the external or internal reference

#         Output:
#             None
#         """
#         logging.debug(__name__ + " :  Set the clock")

#         if val.lower() == "external" or val.lower() == "internal":
#             self._visainstrument.write("source:roscillator:source " + str(val))
#         else:
#             raise ValueError('The reference has to be either "Internal" or "external"')

#     def do_get_reference(self):
#         """
#         Get the reference

#         Input:
#             None

#         Output:
#             state (string) : reference
#         """
#         logging.debug(__name__ + " :  Get the reference")
#         return self._visainstrument.query("source:roscillator:source?")
#         # if self._visainstrument.query('source:roscillator:source?') == 'EXT':
#         #     return 'External'
#         # else:
#         #     return 'Internal'

#     ###################################################################
#     #
#     #
#     #                   Waveform
#     #
#     #
#     ####################################################################

#     def set_waveform_ch1(self, wave):
#         """
#         Sends a complete waveform.

#         Input:
#             w (float[numpoints]) : waveform in V

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : set waveform")

#         # We check if the waveform if no to long
#         if len(wave) > self.maxpoint or len(wave) < 2:

#             logging.debug(
#                 __name__ + " :The waveform is too long or too short (number of points)."
#             )
#             return "Errors : The waveform is too long or too short (number of points)."

#         # We check if the waveform is feasible
#         if (
#             max(wave) > self.max_amplitude / 2 + self.max_offset
#             or min(wave) < -self.max_amplitude / 2 - self.max_offset
#             or max(wave) - min(wave) > self.max_amplitude
#         ):

#             logging.debug(__name__ + " : The waveform is out of the range.")
#             return "Errors : The waveform is out of the range."

#         # We calculate the amplitude and the offset of the tektro
#         amplitude = max(wave) - min(wave)
#         offset = (max(wave) + min(wave)) / 2

#         # We tune these values
#         self.set_amplitude_ch1(amplitude)
#         self.set_offset_ch1(offset)

#         # We transform all points of the waveforme which are in V in bytes
#         wave_bytes = ""
#         for i in wave:

#             # First we tranform V into bins
#             j = int(((i + abs(min(wave))) * (2**14 - 1)) / amplitude)

#             wave_bytes = wave_bytes + struct.pack(">H", j)

#         # We research the beginning of the message for the tektro
#         number_bytes = int(len(wave)) * 2
#         number_digits = len(str(number_bytes))

#         # We create the final message for the device
#         mes = "trace ememory,#" + str(number_digits) + str(number_bytes) + wave_bytes

#         # we prepare the device to be configured
#         self.set_arbitrary_waveform_memory_ch1()
#         self._visainstrument.write(mes)

#     def do_set_function_ch1(self, function="sin"):
#         """
#         Set the function of the channel 1

#         Input:
#             function (string) : function of the channel 1

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the function of the channel 1")
#         self._visainstrument.write("source1:function " + str(function))

#     def do_set_width_ch1(self, width=0.5):
#         """
#         Set the duty cycle of the pulse

#         Input:
#             duty_cycle (float) : duty cycle in seconds

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the duty cycle of the pulse")
#         self._visainstrument.write(f"SOURce1:PULSe:hold width")
#         self._visainstrument.write(f"SOURce1:PULSe:width {width}ns")

#     def do_set_high_level(self, level=0.5):
#         """
#         Set the high level of the pulse

#         Input:
#             level (float) : high level in seconds

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the high level of the pulse")
#         assert level >= 0, "Level must be non-negative"
#         assert level <= 5, "Level must be less than or equal to 5"
#         self._visainstrument.write(f"SOURce1:voltage:level:HIGH {level}")

#     def do_set_level_low(self, level=0.5):
#         """
#         Set the low level of the pulse

#         Input:
#             level (float) : low level in seconds

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the low level of the pulse")
#         assert level >= 0, "Level must be non-negative"
#         assert level <= 5, "Level must be less than or equal to 5"
#         self._visainstrument.write(f"SOURce1:voltage:level:LOW {level}")

#     def do_set_voltage_ch1(self, voltage=0.5):
#         """
#         Set the voltage of the channel 1

#         Input:
#             voltage (float) : voltage of the channel 1

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the voltage of the channel 1")
#         self._visainstrument.write(f"SOURce1:voltage:level:IMMediate:HIGH {voltage}")
#         self._visainstrument.write(f"SOURce1:voltage:level:IMMediate:LOW {0}")

#     def do_set_delay_ch1(self, delay=20):
#         """
#         Set the delay of the pulse

#         Input:
#             delay (float) : delay in seconds

#         Output:
#             None
#         """
#         logging.debug(__name__ + " : Set the delay of the pulse")
#         self._visainstrument.write(f"SOURce1:PULSe:DELay {delay}ns")
